import Editor from '@modules/editor';
import { IObjectClickService, IOrbitControlService, ITransformControlService } from '@modules/editor/service/control';
import { IGNORE_CLICK_TARGET } from '@modules/editor/constants/name';
import * as THREE from 'three';
import { TransformControls } from 'three/examples/jsm/controls/TransformControls.js';

export type Services = {
  TransformControlService: ITransformControlService;
  ObjectClickService: IObjectClickService;
  OrbitControlService: IOrbitControlService;
};

const ignoreClickTargetName: string[] = [IGNORE_CLICK_TARGET.GRID_HELPER_GROUP, IGNORE_CLICK_TARGET.TRANSFORM_CONTROLS];
const ignoreTransformTarget: string[] = ['Line2'];

class EditorControl {
  #Editor: Editor;
  #Services: Services;
  //state
  #isObjectDragging = false;
  #orbitControlOnOff = true;

  constructor(editor: Editor, services: Services) {
    this.#Editor = editor;
    this.#Services = services;
  }

  private set isObjectDragging(value: boolean) {
    this.#isObjectDragging = value;
  }

  private set orbitControlOnOff(onOff: boolean) {
    this.#orbitControlOnOff = onOff;
    this.#Services.OrbitControlService.orbitControls.enabled = onOff;
  }

  static #canvasClickEventCallback(editorControl: EditorControl, e: MouseEvent) {
    const validObj = editorControl.#getClickedValidObject(e);

    if (!validObj) {
      !editorControl.#isObjectDragging && editorControl.detachTransform();
      return;
    }

    editorControl.#Editor.trigger('object_click', { type: 'object_click', target: validObj });
    editorControl.attachTransform(validObj);
  }

  public init() {
    this.turnOnObjectClickEvent();
  }

  public turnOffObjectClickEvent() {
    const transformControl = this.#Services.TransformControlService;
    const { canvas } = this.#Editor;
    transformControl.controller.detach();

    this.#Services.ObjectClickService.removeCanvasClickEvent(canvas, this.#canvasEventCallback);
  }

  public turnOnObjectClickEvent() {
    this.#setCanvasClickEvent();
    this.#setTransformControllerEvent();
  }

  public attachTransform(validObj: THREE.Object3D<THREE.Event>) {
    const transformControl = this.#Services.TransformControlService;
    const isObjectDragging = this.#isObjectDragging;

    const target = ignoreTransformTarget.find((target) => validObj.type === target);
    if (target) return;

    transformControl.attachObjectToControl(isObjectDragging, validObj);
  }

  public detachTransform() {
    const transformControl = this.#Services.TransformControlService;
    transformControl.controller.detach();
  }

  public setTransformMode(mode: 'translate' | 'rotate' | 'scale') {
    const transformControl = this.#Services.TransformControlService;
    if (!transformControl) throw 'error';

    transformControl.controller.setMode(mode);
  }

  public removeClickedObject() {
    const targetObject = this.#Services.TransformControlService.controller.object;
    if (!targetObject) return;

    targetObject.removeFromParent();
    this.#Services.TransformControlService.controller.detach();
  }

  #canvasEventCallback = (e: MouseEvent) => {
    EditorControl.#canvasClickEventCallback(this, e);
  };

  #getClickedValidObject(e: MouseEvent) {
    const objectClick = this.#Services.ObjectClickService;
    const { camera, canvas, scene } = this.#Editor;
    const isObjectDragging = this.#isObjectDragging;

    const validObj = objectClick.getClickObjectEvent(e, {
      scene,
      camera,
      canvas,
      isDragging: isObjectDragging,
      ignoreTargetName: ignoreClickTargetName,
    });

    return validObj;
  }

  #setCanvasClickEvent() {
    const { canvas } = this.#Editor;
    const objectClick = this.#Services.ObjectClickService;

    objectClick.setCanvasClickEvent(canvas, this.#canvasEventCallback);
  }

  #setTransformControllerEvent() {
    const transformControl = this.#Services.TransformControlService;

    transformControl.controller.addEventListener('dragging-changed', (event) => {
      this.isObjectDragging = event.value;
      this.orbitControlOnOff = !event.value;
    });

    transformControl.controller.addEventListener('mouseUp', (e) => {
      const target = e.target as TransformControls;

      if (!target.object) return;

      this.#Editor.EditorStorage.patch({ transform: target.object?.matrixWorld }, target.object.uuid);
    });
  }
}

export default EditorControl;
