import Editor from '@modules/editor/index';
import { Scene } from 'three';
import { getFilesFromItemList } from '@modules/editor/utils';
import { IObjectLoader } from '@modules/editor/service/loader';
import { IFileUploadService } from '@modules/editor/service/upload';

type Services = {
  ObjectLoader: IObjectLoader;
  FileUploadService: IFileUploadService;
};

class EditorUpload {
  public readonly scene: Scene;
  #Editor: Editor;
  #Services: Services;

  constructor(editor: Editor, services: Services) {
    this.#Editor = editor;
    this.#Services = services;
    this.scene = editor.scene;
  }

  public init() {
    this.#setDragDropUploadEvent();
    this.#setDragOverEvent();
  }

  #setDragDropUploadEvent() {
    const { scene } = this.#Editor;
    const editor = this.#Editor;
    const fileUploadService = this.#Services.FileUploadService;

    document.addEventListener('drop', async (event) => {
      event.preventDefault();

      if (!event.dataTransfer || !event.dataTransfer.items) return;

      try {
        fileUploadService.checkValidFile(event);

        editor.EditorEvent.eventTrigger('load_start', {
          type: 'folder_drop',
          target: editor,
          loadStartData: event.dataTransfer.items,
        });

        editor.trigger('load_start', { type: 'folder_drop', target: editor, loadStartData: event.dataTransfer.items });

        const { files } = await fileUploadService.getFilesFromItemList(event.dataTransfer.items);
        const object = await this.#fileLoad(files);
        scene.add(object);

        editor.trigger('load', { type: 'folder_drop', target: editor, id: object.uuid, loadData: files });
        editor.EditorStorage.save(files, object.uuid);
      } catch (e) {
        console.log(e);
      }
    });
  }

  async #fileLoad(files: File[]) {
    const objectGroup = await this.#Services.ObjectLoader.objectLoadAsFolder(files);
    if (!objectGroup) {
      this.#Editor.trigger('load_fail', { type: 'folder_drop', target: this.#Editor });
      return;
    }

    const objectMesh = objectGroup.children[0];

    return objectMesh;
  }

  #setDragOverEvent() {
    document.addEventListener('dragover', (event) => {
      event.preventDefault();

      if (event.dataTransfer) event.dataTransfer.dropEffect = 'copy';
    });
  }
}

export default EditorUpload;
