import * as THREE from 'three';

import EditorScene from '@modules/editor/domain/Editor.Scene';
import EditorCamera from '@modules/editor/domain/Editor.Camera';
import EditorRenderer from '@modules/editor/domain/Editor.Renderer';
import EditorEvent from '@modules/editor/domain/Editor.Event';

import Component from '@egjs/component';
import { Events } from '@modules/editor/type/events';
import EditorControl from '@modules/editor/domain/Editor.Control';
import EditorStorage from '@modules/editor/domain/Editor.Storage';
import EditorUpload from '@modules/editor/domain/Editor.Upload';
import EditorLight from '@modules/editor/domain/Editor.Light';
import EditorGeometry from '@modules/editor/domain/Editor.Geometry';

import { BasePluginType } from '@modules/editor/plugin/plugin.type';
import { GeometryService, GridService } from '@modules/editor/service/geometry';
import { ResizeService } from '@modules/editor/service/resize';
import { IndexedDBService } from '@modules/editor/service/storage';
import { ObjectLoaderService } from '@modules/editor/service/loader';
import FileUploadService from '@modules/editor/service/upload/FileUploadService';
import { OrbitControlService, TransformControlService } from '@modules/editor/service/control';
import ObjectClickService from '@modules/editor/service/control/ObjectClickService';

interface Props {
  canvas: HTMLCanvasElement;
}

class Editor extends Component<Events> {
  //core components
  public readonly EditorCamera: EditorCamera;
  public readonly EditorScene: EditorScene;
  public readonly EditorRenderer: EditorRenderer;
  public readonly EditorEvent: EditorEvent;
  public readonly EditorControl: EditorControl;
  public readonly EditorStorage: EditorStorage;
  public readonly EditorUpload: EditorUpload;
  public readonly EditorLight: EditorLight;
  public readonly EditorGeometry: EditorGeometry;

  //staff
  public readonly canvas: HTMLCanvasElement;
  public readonly camera: THREE.PerspectiveCamera;
  public readonly renderer: THREE.WebGLRenderer;
  public readonly scene: THREE.Scene;
  public readonly plugins: BasePluginType[] = [];

  constructor({ canvas }: Props) {
    super();
    this.canvas = canvas;

    // setting core
    const editorCamera = (this.EditorCamera = new EditorCamera(this));
    const editorScene = (this.EditorScene = new EditorScene(this));
    const editorRenderer = (this.EditorRenderer = new EditorRenderer(this, { ResizeService: new ResizeService() }));

    const scene = (this.scene = editorScene.scene);
    const camera = (this.camera = editorCamera.camera);
    const renderer = (this.renderer = editorRenderer.renderer);

    const editorEvent = (this.EditorEvent = new EditorEvent(this));
    const editorControl = (this.EditorControl = new EditorControl(this, {
      TransformControlService: new TransformControlService({ camera, canvas, scene }),
      OrbitControlService: new OrbitControlService({ camera, canvas }),
      ObjectClickService: new ObjectClickService(),
    }));
    const editorStorage = (this.EditorStorage = new EditorStorage(this, { IndexedDBService: new IndexedDBService() }));
    const editorUpload = (this.EditorUpload = new EditorUpload(this, {
      ObjectLoader: new ObjectLoaderService(),
      FileUploadService: new FileUploadService(),
    }));
    const editorLight = (this.EditorLight = new EditorLight(this));
    const editorGeometry = (this.EditorGeometry = new EditorGeometry(this, {
      GeometryService: new GeometryService(),
      GridService: new GridService(),
    }));
  }

  public async init() {
    this.EditorCamera.init();
    this.EditorScene.init();
    this.EditorRenderer.init();
    this.EditorEvent.init();
    this.EditorControl.init();
    this.EditorStorage.init();
    this.EditorUpload.init();
    this.EditorLight.init();
    this.EditorGeometry.init();

    this.trigger('ready', { type: 'ready', target: this });
    this.trigger('resize', { type: 'resize', target: this });
    this.EditorRenderer.render();
  }

  /**
   * set object controller (transformControls) with translate, rotate, scale.
   * @param mode
   */
  public setTransformMode(mode: 'translate' | 'rotate' | 'scale') {
    return this.EditorControl.setTransformMode(mode);
  }

  /**
   * remove clicked object by transform controller.
   */
  public deleteObject() {
    return this.EditorControl.removeClickedObject();
  }

  /**
   * remove Editor
   */
  public remove() {
    this.trigger('remove', { type: 'remove', target: this });
    this.plugins.forEach((plugin) => {
      plugin.remove();
    });
  }

  public plugin(plugins: BasePluginType[]) {
    plugins.forEach((plugin) => {
      plugin.init();
      this.plugins.push(plugin);
    });
  }
}

export default Editor;
